/*
 * Copyright (c) 2007 Thomas E. Spanjaard <tgen@netphreax.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/callout.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/proc.h>
#include <sys/spinlock.h>
#include <sys/spinlock2.h>
#include <sys/sysctl.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>

/* constants */
#define IA32_THERM_STATUS	0x19c	/* thermal status MSR address */
#define CORETEMP_POLLRATE	10	/* poll every 10 seconds by default */

/* Tj(Max) values determined empirically */
#define E4300_TJMAX		100
#define E6600_TJMAX		85
#define Q6600_TJMAX		100

/* structure definitions */
struct coretemp_pcpu {
	int			temperature;
	struct callout		callout;
	struct sysctl_oid	*oidp;
};

/* local vars */
static int			coretemp_cleanup = 0;
static int			coretemp_tjmax = 0;
static int			coretemp_pollrate = 0;
static struct coretemp_pcpu	coretemp_cpu[32];	/* 32 CPUs max */
static struct spinlock		coretemp_lock;

/* sysctl vars */
static struct sysctl_ctx_list	coretemp_ctx;
static struct sysctl_oid	*coretemp_oid;

/*
 * Module which reads out CPU core temperatures and exports them as sysctls.
 */

static void
coretemp_updtemp(struct coretemp_pcpu *cpu)
{
	u_int64_t		therm_status;
	int			readout, resolution, temperature;

	spin_lock_rd(&coretemp_lock);
	if (coretemp_cleanup != 0) {
		kprintf("coretemp: cpu%d cleaning up\n", mycpuid);
		sysctl_remove_oid(cpu->oidp, 1, 0);
		cpu->temperature = -256;	/* magic ;) */
		callout_stop(&cpu->callout);
		spin_unlock_rd(&coretemp_lock);
		wakeup(&coretemp_cleanup);
		return;
	}
	spin_unlock_rd(&coretemp_lock);

	therm_status = rdmsr(IA32_THERM_STATUS);
	readout = (int)((therm_status >> 16) & 0x7f);
	resolution = (int)((therm_status >> 27) & 0xf);
	temperature = coretemp_tjmax - (readout * resolution);
	cpu->temperature = temperature;
	callout_reset(&cpu->callout, coretemp_pollrate * hz,
		      (timeout_t *)coretemp_updtemp, cpu);
}

static void
coretemp_init(void)
{
	char			oidname[8];
	u_int			regs[4];
	int			cpu, origcpu;

	/* check if CPU supports DTS */
	do_cpuid(0x6, regs);
	if (!((regs[0] & 0x1) == 0x1)) {
		kprintf("coretemp: CPU does not have a DTS\n");
		coretemp_cleanup = 1;
		return;
	}

	/* determine which Tj(Max) to use */
	coretemp_tjmax = Q6600_TJMAX;	/* I'm a lazy bum right now */
	coretemp_pollrate = CORETEMP_POLLRATE;

	/* enumerate all CPUs, create sysctls for them */
	sysctl_ctx_init(&coretemp_ctx);
	coretemp_oid = SYSCTL_ADD_NODE(&coretemp_ctx, SYSCTL_STATIC_CHILDREN(_hw),
				       OID_AUTO, "coretemp", CTLFLAG_RD, 0, "");
	SYSCTL_ADD_INT(&coretemp_ctx, SYSCTL_CHILDREN(coretemp_oid), OID_AUTO,
		       "polling_rate", CTLFLAG_RD | CTLFLAG_RW, &coretemp_pollrate,
		       0, "monitor polling rate");
	spin_init(&coretemp_lock);
	origcpu = mycpuid;
	for (cpu = 0; cpu < ncpus; cpu++) {
		struct sysctl_oid	*oidp;

		lwkt_migratecpu(cpu);

		ksprintf(oidname, "cpu%d", cpu);
		oidp = SYSCTL_ADD_INT(&coretemp_ctx, SYSCTL_CHILDREN(coretemp_oid),
			       OID_AUTO, oidname, CTLFLAG_RD,
			       &coretemp_cpu[cpu].temperature, 0, "");
		coretemp_cpu[cpu].oidp = oidp;
		callout_init_mp(&coretemp_cpu[cpu].callout);
		callout_reset(&coretemp_cpu[cpu].callout, coretemp_pollrate * hz,
			      (timeout_t *)coretemp_updtemp, &coretemp_cpu[cpu]);
	}
	lwkt_migratecpu(origcpu);
}

SYSINIT(coretemp_register, SI_SUB_DRIVERS, SI_ORDER_ANY, coretemp_init, NULL);

static void
coretemp_uninit(void)
{
	int			cpu;

	if (coretemp_cleanup == 1)
		return;

	spin_lock_wr(&coretemp_lock);
	coretemp_cleanup = 1;
	for (cpu = 0; cpu < ncpus; cpu++) {
		while (coretemp_cpu[cpu].temperature != -256) {
			spin_unlock_wr(&coretemp_lock);
			kprintf("coretemp: waiting for cpu%d to clean up\n", cpu);
			tsleep(&coretemp_cleanup, 0, "ctcln", coretemp_pollrate * hz);
			spin_lock_wr(&coretemp_lock);
		}
	}
	sysctl_ctx_free(&coretemp_ctx);
	spin_unlock_wr(&coretemp_lock);
	spin_uninit(&coretemp_lock);
}

SYSUNINIT(coretemp_unregister, SI_SUB_DRIVERS, SI_ORDER_ANY, coretemp_uninit, NULL);

/*
 * Module handling.
 */

static int
coretemp_module_event_handler(module_t mod, int what, void *arg)
{
	switch(what) {
	case MOD_LOAD:
		return 0;

	case MOD_UNLOAD:
		return 0;

	default:
		return EOPNOTSUPP;
	}
}

static moduledata_t coretemp_moduledata = {"coretemp", coretemp_module_event_handler,
					   NULL};
DECLARE_MODULE(coretemp, coretemp_moduledata, SI_SUB_CONFIGURE, SI_ORDER_ANY);
MODULE_VERSION(coretemp, 1);
